using System.Collections;
using System.Diagnostics.Contracts;

namespace System.Threading
{
    // only this one lock is used, so no deadlock may occur
    internal static class BigLock
    {
        internal readonly static object bigLock = new object();
    }


    public delegate void WaitCallback(object state);

    internal class WorkItem
    {
        internal readonly WaitCallback callback;
        internal readonly object state;

        internal WorkItem(WaitCallback callback, object state)
        {
            Contract.Requires(callback != null);
            this.callback = callback;
            this.state = state;
        }
    }

    public static class ThreadPool
    {
        static Thread[] threads;
        static BoundedQueue workItemQueue;

        public static void Open(int nofThreads)
        {
            Contract.Requires(nofThreads > 0);
            threads = new Thread[nofThreads];
            var i = 0;
            while (i != nofThreads)
            {
                threads[i] = new Thread(Run);
                threads[i].Priority = ThreadPriority.Lowest;
                i = i + 1;
            }
            workItemQueue = new BoundedQueue { Capacity = nofThreads };
        }

        public static void Start()
        {
            var i = 0;
            while (i != threads.Length)
            {
                threads[i].Start();
                i = i + 1;
            }
        }

        public static bool QueueUserWorkItem(WaitCallback callback,
                                                object state)
        {
            Contract.Requires(callback != null);
            Contract.Requires(threads != null);
            var item = new WorkItem(callback, state);
            workItemQueue.Enqueue(item);
            return true;
        }

        static void Run()
        {
            while (true)
            {
                var obj = workItemQueue.Dequeue();      // may block!
                var item = (WorkItem)obj;
                try
                {
                    item.callback(item.state);
                }
                catch (Exception e)
                {
                    //Debug.Print("ThreadPool.Run exception in callback:\r\n" + e);
                }
            }
        }
    }


    // TODO documentation
    internal sealed class BoundedQueue
    {
        AutoResetEvent nonEmptySignal = new AutoResetEvent(false);
        AutoResetEvent nonFullSignal = new AutoResetEvent(false);

        int capacity = 0;           // where (capacity >= 0)
        Queue queue = new Queue();  // where (queue != null)

        internal int Capacity         // once set, remains immutable
        {
            get { return capacity; }

            set
            {
                Contract.Requires(value > 0);
                capacity = value;
            }
        }

        internal void Enqueue(object o)
        {
            Contract.Requires(capacity > 0);
            // this method may block
            int count;
            lock (BigLock.bigLock)
            {
                count = queue.Count;
                while (count == capacity)
                {
                    Monitor.Exit(BigLock.bigLock);
                    nonFullSignal.WaitOne();
                    Monitor.Enter(BigLock.bigLock);
                    count = queue.Count;
                }
                queue.Enqueue(o);
            }
            if (count == 0)
            {
                nonEmptySignal.Set();
            }
        }


        internal object Dequeue()
        {
            Contract.Requires(capacity > 0);
            // this method may block
            object o = null;
            int count;
            lock (BigLock.bigLock)
            {
                count = queue.Count;
                while (count == 0)
                {
                    Monitor.Exit(BigLock.bigLock);
                    nonEmptySignal.WaitOne();
                    Monitor.Enter(BigLock.bigLock);
                    count = queue.Count;
                }
                o = queue.Dequeue();
            }
            if (count == capacity)
            {
                nonFullSignal.Set();
            }
            return o;
        }
    }
}
